@@ -1,6 +1,6 @@
-/* the Music Player Daemon (MPD)
- * Copyright (C) 2003-2007 by Warren Dukes (warren.dukes@gmail.com)
- * This project's homepage is: http://www.musicpd.org
+/*
+ * Copyright (C) 2003-2009 The Music Player Daemon Project
+ * http://www.musicpd.org
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -11,15 +11,20 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include "command.h"
 #include "player_control.h"
 #include "playlist.h"
+#include "playlist_print.h"
+#include "playlist_save.h"
+#include "queue_print.h"
 #include "ls.h"
+#include "uri.h"
 #include "directory.h"
 #include "directory_print.h"
 #include "database.h"
@@ -28,27 +33,43 @@
 #include "stats.h"
 #include "permission.h"
 #include "buffer2array.h"
-#include "log.h"
-#include "utils.h"
 #include "stored_playlist.h"
 #include "ack.h"
-#include "audio.h"
+#include "output_command.h"
+#include "output_print.h"
+#include "locate.h"
 #include "dbUtils.h"
 #include "tag.h"
 #include "client.h"
 #include "tag_print.h"
 #include "path.h"
-#include "os_compat.h"
 #include "idle.h"
+#include "config.h"
+
+#ifdef ENABLE_SQLITE
+#include "sticker.h"
+#include "sticker_print.h"
+#include "song_sticker.h"
+#include "song_print.h"
+#endif
+
+#include <assert.h>
+#include <time.h>
+#include <stdlib.h>
+#include <errno.h>
 
 #define COMMAND_STATUS_VOLUME           "volume"
 #define COMMAND_STATUS_STATE            "state"
 #define COMMAND_STATUS_REPEAT           "repeat"
+#define COMMAND_STATUS_SINGLE           "single"
+#define COMMAND_STATUS_CONSUME          "consume"
 #define COMMAND_STATUS_RANDOM           "random"
 #define COMMAND_STATUS_PLAYLIST         "playlist"
 #define COMMAND_STATUS_PLAYLIST_LENGTH  "playlistlength"
 #define COMMAND_STATUS_SONG             "song"
 #define COMMAND_STATUS_SONGID           "songid"
+#define COMMAND_STATUS_NEXTSONG         "nextsong"
+#define COMMAND_STATUS_NEXTSONGID       "nextsongid"
 #define COMMAND_STATUS_TIME             "time"
 #define COMMAND_STATUS_BITRATE          "bitrate"
 #define COMMAND_STATUS_ERROR            "error"
@@ -75,6 +96,7 @@ struct command {
 
 /* this should really be "need a non-negative integer": */
 static const char need_positive[] = "need a positive integer"; /* no-op */
+static const char need_range[] = "need a range";
 
 /* FIXME: redundant error messages */
 static const char check_integer[] = "\"%s\" is not a integer";
@@ -102,7 +124,7 @@ static void command_error_v(struct client *client, enum ack error,
 	current_command = NULL;
 }
 
-G_GNUC_PRINTF(3, 4) void command_error(struct client *client, enum ack error,
+G_GNUC_PRINTF(3, 4) static void command_error(struct client *client, enum ack error,
 				       const char *fmt, ...)
 {
 	va_list args;
@@ -156,6 +178,77 @@ check_int(struct client *client, int *value_r,
 	return true;
 }
 
+static bool G_GNUC_PRINTF(5, 6)
+check_range(struct client *client, unsigned *value_r1, unsigned *value_r2,
+	    const char *s, const char *fmt, ...)
+{
+	char *test, *test2;
+	long value;
+
+	value = strtol(s, &test, 10);
+	if (*test != '\0' && *test != ':') {
+		va_list args;
+		va_start(args, fmt);
+		command_error_v(client, ACK_ERROR_ARG, fmt, args);
+		va_end(args);
+		return false;
+	}
+
+	if (value == -1 && *test == 0) {
+		/* compatibility with older MPD versions: specifying
+		   "-1" makes MPD display the whole list */
+		*value_r1 = 0;
+		*value_r2 = UINT_MAX;
+		return true;
+	}
+
+	if (value < 0) {
+		command_error(client, ACK_ERROR_ARG,
+			      "Number is negative: %s", s);
+		return false;
+	}
+
+#if LONG_MAX > UINT_MAX
+	if (value > UINT_MAX) {
+		command_error(client, ACK_ERROR_ARG,
+			      "Number too large: %s", s);
+		return false;
+	}
+#endif
+
+	*value_r1 = (unsigned)value;
+
+	if (*test == ':') {
+		value = strtol(++test, &test2, 10);
+		if (*test2 != '\0' || test == test2) {
+			va_list args;
+			va_start(args, fmt);
+			command_error_v(client, ACK_ERROR_ARG, fmt, args);
+			va_end(args);
+			return false;
+		}
+
+		if (value < 0) {
+			command_error(client, ACK_ERROR_ARG,
+				      "Number is negative: %s", s);
+			return false;
+		}
+
+#if LONG_MAX > UINT_MAX
+		if (value > UINT_MAX) {
+			command_error(client, ACK_ERROR_ARG,
+				      "Number too large: %s", s);
+			return false;
+		}
+#endif
+		*value_r2 = (unsigned)value;
+	} else {
+		*value_r2 = (unsigned)value + 1;
+	}
+
+	return true;
+}
+
 static bool
 check_unsigned(struct client *client, unsigned *value_r, const char *s)
 {
@@ -244,6 +337,11 @@ print_playlist_result(struct client *client,
 		command_error(client, ACK_ERROR_PLAYLIST_MAX,
 			      "playlist is at the max size");
 		return COMMAND_RETURN_ERROR;
+
+	case PLAYLIST_RESULT_DISABLED:
+		command_error(client, ACK_ERROR_UNKNOWN,
+			      "stored playlist support is disabled");
+		return COMMAND_RETURN_ERROR;
 	}
 
 	assert(0);
@@ -257,14 +355,21 @@ print_spl_list(struct client *client, GPtrArray *list)
 		struct stored_playlist_info *playlist =
 			g_ptr_array_index(list, i);
 		time_t t;
+#ifndef WIN32
 		struct tm tm;
+#endif
 		char timestamp[32];
 
 		client_printf(client, "playlist: %s\n", playlist->name);
 
 		t = playlist->mtime;
 		strftime(timestamp, sizeof(timestamp), "%FT%TZ",
-			 gmtime_r(&t, &tm));
+#ifdef WIN32
+			 gmtime(&t)
+#else
+			 gmtime_r(&t, &tm)
+#endif
+			 );
 		client_printf(client, "Last-Modified: %s\n", timestamp);
 	}
 }
@@ -275,7 +380,7 @@ handle_urlhandlers(struct client *client,
 {
 	if (client_get_uid(client) > 0)
 		client_puts(client, "handler: file://\n");
-	printRemoteUrlHandlers(client);
+	print_supported_uri_schemes(client);
 	return COMMAND_RETURN_OK;
 }
 
@@ -295,7 +400,7 @@ handle_play(struct client *client, int argc, char *argv[])
 
 	if (argc == 2 && !check_int(client, &song, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
-	result = playPlaylist(song, 0);
+	result = playPlaylist(&g_playlist, song);
 	return print_playlist_result(client, result);
 }
 
@@ -308,7 +413,7 @@ handle_playid(struct client *client, int argc, char *argv[])
 	if (argc == 2 && !check_int(client, &id, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
 
-	result = playPlaylistById(id, 0);
+	result = playPlaylistById(&g_playlist, id);
 	return print_playlist_result(client, result);
 }
 
@@ -316,7 +421,7 @@ static enum command_return
 handle_stop(G_GNUC_UNUSED struct client *client,
 	    G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	stopPlaylist();
+	stopPlaylist(&g_playlist);
 	return COMMAND_RETURN_OK;
 }
 
@@ -324,14 +429,8 @@ static enum command_return
 handle_currentsong(struct client *client,
 		   G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	int song = getPlaylistCurrentSong();
-	enum playlist_result result;
-
-	if (song < 0)
-		return COMMAND_RETURN_OK;
-
-	result = playlistInfo(client, song);
-	return print_playlist_result(client, result);
+	playlist_print_current(client, &g_playlist);
+	return PLAYLIST_RESULT_SUCCESS;
 }
 
 static enum command_return
@@ -358,7 +457,6 @@ handle_status(struct client *client,
 	int updateJobId;
 	int song;
 
-	playPlaylistIfPlayerStopped();
 	switch (getPlayerState()) {
 	case PLAYER_STATE_STOP:
 		state = "stop";
@@ -375,24 +473,28 @@ handle_status(struct client *client,
 		      COMMAND_STATUS_VOLUME ": %i\n"
 		      COMMAND_STATUS_REPEAT ": %i\n"
 		      COMMAND_STATUS_RANDOM ": %i\n"
+		      COMMAND_STATUS_SINGLE ": %i\n"
+		      COMMAND_STATUS_CONSUME ": %i\n"
 		      COMMAND_STATUS_PLAYLIST ": %li\n"
 		      COMMAND_STATUS_PLAYLIST_LENGTH ": %i\n"
 		      COMMAND_STATUS_CROSSFADE ": %i\n"
 		      COMMAND_STATUS_STATE ": %s\n",
-		      getVolumeLevel(),
-		      getPlaylistRepeatStatus(),
-		      getPlaylistRandomStatus(),
-		      getPlaylistVersion(),
-		      getPlaylistLength(),
+		      volume_level_get(),
+		      getPlaylistRepeatStatus(&g_playlist),
+		      getPlaylistRandomStatus(&g_playlist),
+		      getPlaylistSingleStatus(&g_playlist),
+		      getPlaylistConsumeStatus(&g_playlist),
+		      getPlaylistVersion(&g_playlist),
+		      getPlaylistLength(&g_playlist),
 		      (int)(getPlayerCrossFade() + 0.5),
 		      state);
 
-	song = getPlaylistCurrentSong();
+	song = getPlaylistCurrentSong(&g_playlist);
 	if (song >= 0) {
 		client_printf(client,
 			      COMMAND_STATUS_SONG ": %i\n"
 			      COMMAND_STATUS_SONGID ": %u\n",
-			      song, getPlaylistSongId(song));
+			      song, getPlaylistSongId(&g_playlist, song));
 	}
 
 	if (getPlayerState() != PLAYER_STATE_STOP) {
@@ -418,6 +520,14 @@ handle_status(struct client *client,
 			      getPlayerErrorStr());
 	}
 
+	song = getPlaylistNextSong(&g_playlist);
+	if (song >= 0) {
+		client_printf(client,
+			      COMMAND_STATUS_NEXTSONG ": %i\n"
+			      COMMAND_STATUS_NEXTSONGID ": %u\n",
+			      song, getPlaylistSongId(&g_playlist, song));
+	}
+
 	return COMMAND_RETURN_OK;
 }
 
@@ -438,25 +548,32 @@ handle_close(G_GNUC_UNUSED struct client *client,
 static enum command_return
 handle_add(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
-	char *path = argv[1];
+	char *uri = argv[1];
 	enum playlist_result result;
 
-	if (strncmp(path, "file:///", 8) == 0) {
-		result = playlist_append_file(path + 7, client_get_uid(client),
+	if (strncmp(uri, "file:///", 8) == 0) {
+#ifdef WIN32
+		result = PLAYLIST_RESULT_DENIED;
+#else
+		result = playlist_append_file(&g_playlist,
+					      uri + 7, client_get_uid(client),
 					      NULL);
+#endif
 		return print_playlist_result(client, result);
 	}
 
-	if (isRemoteUrl(path))
-		return addToPlaylist(path, NULL);
+	if (uri_has_scheme(uri)) {
+		if (!uri_supported_scheme(uri)) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "unsupported URI scheme");
+			return COMMAND_RETURN_ERROR;
+		}
 
-	if (uri_has_scheme(path)) {
-		command_error(client, ACK_ERROR_NO_EXIST,
-			      "unsupported URI scheme");
-		return COMMAND_RETURN_ERROR;
+		result = addToPlaylist(&g_playlist, uri, NULL);
+		return print_playlist_result(client, result);
 	}
 
-	result = addAllIn(path);
+	result = addAllIn(uri);
 	if (result == (enum playlist_result)-1) {
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "directory or file not found");
@@ -469,15 +586,27 @@ handle_add(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 static enum command_return
 handle_addid(struct client *client, int argc, char *argv[])
 {
+	char *uri = argv[1];
 	unsigned added_id;
 	enum playlist_result result;
 
-	if (strncmp(argv[1], "file:///", 8) == 0)
-		result = playlist_append_file(argv[1] + 7,
+	if (strncmp(uri, "file:///", 8) == 0) {
+#ifdef WIN32
+		result = PLAYLIST_RESULT_DENIED;
+#else
+		result = playlist_append_file(&g_playlist, uri + 7,
 					      client_get_uid(client),
 					      &added_id);
-	else
-		result = addToPlaylist(argv[1], &added_id);
+#endif
+	} else {
+		if (uri_has_scheme(uri) && !uri_supported_scheme(uri)) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "unsupported URI scheme");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		result = addToPlaylist(&g_playlist, uri, &added_id);
+	}
 
 	if (result != PLAYLIST_RESULT_SUCCESS)
 		return print_playlist_result(client, result);
@@ -486,17 +615,17 @@ handle_addid(struct client *client, int argc, char *argv[])
 		int to;
 		if (!check_int(client, &to, argv[2], check_integer, argv[2]))
 			return COMMAND_RETURN_ERROR;
-		result = moveSongInPlaylistById(added_id, to);
+		result = moveSongInPlaylistById(&g_playlist, added_id, to);
 		if (result != PLAYLIST_RESULT_SUCCESS) {
 			enum command_return ret =
 				print_playlist_result(client, result);
-			deleteFromPlaylistById(added_id);
+			deleteFromPlaylistById(&g_playlist, added_id);
 			return ret;
 		}
 	}
 
 	client_printf(client, "Id: %u\n", added_id);
-	return result;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
@@ -508,7 +637,7 @@ handle_delete(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 	if (!check_int(client, &song, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
 
-	result = deleteFromPlaylist(song);
+	result = deleteFromPlaylist(&g_playlist, song);
 	return print_playlist_result(client, result);
 }
 
@@ -521,7 +650,7 @@ handle_deleteid(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 	if (!check_int(client, &id, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
 
-	result = deleteFromPlaylistById(id);
+	result = deleteFromPlaylistById(&g_playlist, id);
 	return print_playlist_result(client, result);
 }
 
@@ -529,7 +658,7 @@ static enum command_return
 handle_playlist(struct client *client,
 	        G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	showPlaylist(client);
+	playlist_print_uris(client, &g_playlist);
 	return COMMAND_RETURN_OK;
 }
 
@@ -537,7 +666,12 @@ static enum command_return
 handle_shuffle(G_GNUC_UNUSED struct client *client,
 	       G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	shufflePlaylist();
+	unsigned start = 0, end = queue_length(&g_playlist.queue);
+	if (argc == 2 && !check_range(client, &start, &end,
+	                              argv[1], need_range))
+		return COMMAND_RETURN_ERROR;
+
+	shufflePlaylist(&g_playlist, start, end);
 	return COMMAND_RETURN_OK;
 }
 
@@ -545,7 +679,7 @@ static enum command_return
 handle_clear(G_GNUC_UNUSED struct client *client,
 	     G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	clearPlaylist();
+	clearPlaylist(&g_playlist);
 	return COMMAND_RETURN_OK;
 }
 
@@ -555,7 +689,7 @@ handle_save(struct client *client,
 {
 	enum playlist_result result;
 
-	result = savePlaylist(argv[1]);
+	result = spl_save_playlist(argv[1], &g_playlist);
 	return print_playlist_result(client, result);
 }
 
@@ -564,45 +698,52 @@ handle_load(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
 	enum playlist_result result;
 
-	result = loadPlaylist(argv[1]);
+	result = playlist_load_spl(&g_playlist, argv[1]);
 	return print_playlist_result(client, result);
 }
 
 static enum command_return
 handle_listplaylist(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
-	int ret;
+	bool ret;
 
-	ret = PlaylistInfo(client, argv[1], 0);
-	if (ret == -1)
+	ret = spl_print(client, argv[1], false);
+	if (!ret) {
 		command_error(client, ACK_ERROR_NO_EXIST, "No such playlist");
+		return COMMAND_RETURN_ERROR;
+	}
 
-	return ret;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_listplaylistinfo(struct client *client,
 			G_GNUC_UNUSED int argc, char *argv[])
 {
-	int ret;
+	bool ret;
 
-	ret = PlaylistInfo(client, argv[1], 1);
-	if (ret == -1)
+	ret = spl_print(client, argv[1], true);
+	if (!ret) {
 		command_error(client, ACK_ERROR_NO_EXIST, "No such playlist");
+		return COMMAND_RETURN_ERROR;
+	}
 
-	return ret;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_lsinfo(struct client *client, int argc, char *argv[])
 {
-	const char *path = "";
+	const char *uri;
 	const struct directory *directory;
 
 	if (argc == 2)
-		path = argv[1];
+		uri = argv[1];
+	else
+		/* default is root directory */
+		uri = "";
 
-	directory = db_get_directory(path);
+	directory = db_get_directory(uri);
 	if (directory == NULL) {
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "directory not found");
@@ -611,7 +752,7 @@ handle_lsinfo(struct client *client, int argc, char *argv[])
 
 	directory_print(client, directory);
 
-	if (isRootDirectory(path)) {
+	if (isRootDirectory(uri)) {
 		GPtrArray *list = spl_list();
 		if (list != NULL) {
 			print_spl_list(client, list);
@@ -647,7 +788,9 @@ handle_plchanges(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 
 	if (!check_uint32(client, &version, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
-	return playlistChanges(client, version);
+
+	playlist_print_changes_info(client, &g_playlist, version);
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
@@ -657,56 +800,70 @@ handle_plchangesposid(struct client *client, G_GNUC_UNUSED int argc, char *argv[
 
 	if (!check_uint32(client, &version, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
-	return playlistChangesPosId(client, version);
+
+	playlist_print_changes_position(client, &g_playlist, version);
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_playlistinfo(struct client *client, int argc, char *argv[])
 {
-	int song = -1;
-	enum playlist_result result;
+	unsigned start = 0, end = UINT_MAX;
+	bool ret;
 
-	if (argc == 2 && !check_int(client, &song, argv[1], need_positive))
+	if (argc == 2 && !check_range(client, &start, &end,
+				      argv[1], need_range))
 		return COMMAND_RETURN_ERROR;
 
-	result = playlistInfo(client, song);
-	return print_playlist_result(client, result);
+	ret = playlist_print_info(client, &g_playlist, start, end);
+	if (!ret)
+		return print_playlist_result(client,
+					     PLAYLIST_RESULT_BAD_RANGE);
+
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_playlistid(struct client *client, int argc, char *argv[])
 {
 	int id = -1;
-	enum playlist_result result;
 
 	if (argc == 2 && !check_int(client, &id, argv[1], need_positive))
 		return COMMAND_RETURN_ERROR;
 
-	result = playlistId(client, id);
-	return print_playlist_result(client, result);
+	if (id >= 0) {
+		bool ret = playlist_print_id(client, &g_playlist, id);
+		if (!ret)
+			return print_playlist_result(client,
+						     PLAYLIST_RESULT_NO_SUCH_SONG);
+	} else {
+		playlist_print_info(client, &g_playlist, 0, UINT_MAX);
+	}
+
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_find(struct client *client, int argc, char *argv[])
 {
 	int ret;
+	struct locate_item_list *list =
+		locate_item_list_parse(argv + 1, argc - 1);
 
-	LocateTagItem *items;
-	int numItems = newLocateTagItemArrayFromArgArray(argv + 1,
-							 argc - 1,
-							 &items);
+	if (list == NULL || list->length == 0) {
+		if (list != NULL)
+			locate_item_list_free(list);
 
-	if (numItems <= 0) {
 		command_error(client, ACK_ERROR_ARG, "incorrect arguments");
 		return COMMAND_RETURN_ERROR;
 	}
 
-	ret = findSongsIn(client, NULL, numItems, items);
+	ret = findSongsIn(client, NULL, list);
 	if (ret == -1)
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "directory or file not found");
 
-	freeLocateTagItemArray(numItems, items);
+	locate_item_list_free(list);
 
 	return ret;
 }
@@ -715,23 +872,23 @@ static enum command_return
 handle_search(struct client *client, int argc, char *argv[])
 {
 	int ret;
+	struct locate_item_list *list =
+		locate_item_list_parse(argv + 1, argc - 1);
 
-	LocateTagItem *items;
-	int numItems = newLocateTagItemArrayFromArgArray(argv + 1,
-							 argc - 1,
-							 &items);
+	if (list == NULL || list->length == 0) {
+		if (list != NULL)
+			locate_item_list_free(list);
 
-	if (numItems <= 0) {
 		command_error(client, ACK_ERROR_ARG, "incorrect arguments");
 		return COMMAND_RETURN_ERROR;
 	}
 
-	ret = searchForSongsIn(client, NULL, numItems, items);
+	ret = searchForSongsIn(client, NULL, list);
 	if (ret == -1)
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "directory or file not found");
 
-	freeLocateTagItemArray(numItems, items);
+	locate_item_list_free(list);
 
 	return ret;
 }
@@ -740,23 +897,23 @@ static enum command_return
 handle_count(struct client *client, int argc, char *argv[])
 {
 	int ret;
+	struct locate_item_list *list =
+		locate_item_list_parse(argv + 1, argc - 1);
 
-	LocateTagItem *items;
-	int numItems = newLocateTagItemArrayFromArgArray(argv + 1,
-							 argc - 1,
-							 &items);
+	if (list == NULL || list->length == 0) {
+		if (list != NULL)
+			locate_item_list_free(list);
 
-	if (numItems <= 0) {
 		command_error(client, ACK_ERROR_ARG, "incorrect arguments");
 		return COMMAND_RETURN_ERROR;
 	}
 
-	ret = searchStatsForSongsIn(client, NULL, numItems, items);
+	ret = searchStatsForSongsIn(client, NULL, list);
 	if (ret == -1)
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "directory or file not found");
 
-	freeLocateTagItemArray(numItems, items);
+	locate_item_list_free(list);
 
 	return ret;
 }
@@ -764,19 +921,20 @@ handle_count(struct client *client, int argc, char *argv[])
 static enum command_return
 handle_playlistfind(struct client *client, int argc, char *argv[])
 {
-	LocateTagItem *items;
-	int numItems = newLocateTagItemArrayFromArgArray(argv + 1,
-							 argc - 1,
-							 &items);
+	struct locate_item_list *list =
+		locate_item_list_parse(argv + 1, argc - 1);
+
+	if (list == NULL || list->length == 0) {
+		if (list != NULL)
+			locate_item_list_free(list);
 
-	if (numItems <= 0) {
 		command_error(client, ACK_ERROR_ARG, "incorrect arguments");
 		return COMMAND_RETURN_ERROR;
 	}
 
-	findSongsInPlaylist(client, numItems, items);
+	playlist_print_find(client, &g_playlist, list);
 
-	freeLocateTagItemArray(numItems, items);
+	locate_item_list_free(list);
 
 	return COMMAND_RETURN_OK;
 }
@@ -784,19 +942,20 @@ handle_playlistfind(struct client *client, int argc, char *argv[])
 static enum command_return
 handle_playlistsearch(struct client *client, int argc, char *argv[])
 {
-	LocateTagItem *items;
-	int numItems = newLocateTagItemArrayFromArgArray(argv + 1,
-							 argc - 1,
-							 &items);
+	struct locate_item_list *list =
+		locate_item_list_parse(argv + 1, argc - 1);
+
+	if (list == NULL || list->length == 0) {
+		if (list != NULL)
+			locate_item_list_free(list);
 
-	if (numItems <= 0) {
 		command_error(client, ACK_ERROR_ARG, "incorrect arguments");
 		return COMMAND_RETURN_ERROR;
 	}
 
-	searchForSongsInPlaylist(client, numItems, items);
+	playlist_print_search(client, &g_playlist, list);
 
-	freeLocateTagItemArray(numItems, items);
+	locate_item_list_free(list);
 
 	return COMMAND_RETURN_OK;
 }
@@ -856,7 +1015,14 @@ static enum command_return
 handle_next(G_GNUC_UNUSED struct client *client,
 	    G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	nextSongInPlaylist();
+	/* single mode is not considered when this is user who
+	 * wants to change song. */
+	int single = g_playlist.queue.single;
+	g_playlist.queue.single = false;
+
+	nextSongInPlaylist(&g_playlist);
+
+	g_playlist.queue.single = single;
 	return COMMAND_RETURN_OK;
 }
 
@@ -864,7 +1030,7 @@ static enum command_return
 handle_previous(G_GNUC_UNUSED struct client *client,
 		G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	previousSongInPlaylist();
+	previousSongInPlaylist(&g_playlist);
 	return COMMAND_RETURN_OK;
 }
 
@@ -888,33 +1054,39 @@ handle_listall(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 static enum command_return
 handle_volume(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
-	int change, ret;
+	int change;
+	bool success;
 
 	if (!check_int(client, &change, argv[1], need_integer))
 		return COMMAND_RETURN_ERROR;
 
-	ret = changeVolumeLevel(change, 1);
-	if (ret == -1)
+	success = volume_level_change(change, true);
+	if (!success) {
 		command_error(client, ACK_ERROR_SYSTEM,
 			      "problems setting volume");
+		return COMMAND_RETURN_ERROR;
+	}
 
-	return ret;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_setvol(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
-	int level, ret;
+	int level;
+	bool success;
 
 	if (!check_int(client, &level, argv[1], need_integer))
 		return COMMAND_RETURN_ERROR;
 
-	ret = changeVolumeLevel(level, 0);
-	if (ret == -1)
+	success = volume_level_change(level, 0);
+	if (!success) {
 		command_error(client, ACK_ERROR_SYSTEM,
 			      "problems setting volume");
+		return COMMAND_RETURN_ERROR;
+	}
 
-	return ret;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
@@ -931,7 +1103,43 @@ handle_repeat(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 		return COMMAND_RETURN_ERROR;
 	}
 
-	setPlaylistRepeatStatus(status);
+	setPlaylistRepeatStatus(&g_playlist, status);
+	return COMMAND_RETURN_OK;
+}
+
+static enum command_return
+handle_single(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
+{
+	int status;
+
+	if (!check_int(client, &status, argv[1], need_integer))
+		return COMMAND_RETURN_ERROR;
+
+	if (status != 0 && status != 1) {
+		command_error(client, ACK_ERROR_ARG,
+			      "\"%i\" is not 0 or 1", status);
+		return COMMAND_RETURN_ERROR;
+	}
+
+	setPlaylistSingleStatus(&g_playlist, status);
+	return COMMAND_RETURN_OK;
+}
+
+static enum command_return
+handle_consume(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
+{
+	int status;
+
+	if (!check_int(client, &status, argv[1], need_integer))
+		return COMMAND_RETURN_ERROR;
+
+	if (status != 0 && status != 1) {
+		command_error(client, ACK_ERROR_ARG,
+			      "\"%i\" is not 0 or 1", status);
+		return COMMAND_RETURN_ERROR;
+	}
+
+	setPlaylistConsumeStatus(&g_playlist, status);
 	return COMMAND_RETURN_OK;
 }
 
@@ -949,7 +1157,7 @@ handle_random(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 		return COMMAND_RETURN_ERROR;
 	}
 
-	setPlaylistRandomStatus(status);
+	setPlaylistRandomStatus(&g_playlist, status);
 	return COMMAND_RETURN_OK;
 }
 
@@ -957,7 +1165,7 @@ static enum command_return
 handle_stats(struct client *client,
 	     G_GNUC_UNUSED int argc, G_GNUC_UNUSED char *argv[])
 {
-	return printStats(client);
+	return stats_print(client);
 }
 
 static enum command_return
@@ -971,9 +1179,8 @@ handle_clearerror(G_GNUC_UNUSED struct client *client,
 static enum command_return
 handle_list(struct client *client, int argc, char *argv[])
 {
-	int numConditionals;
-	LocateTagItem *conditionals = NULL;
-	int tagType = getLocateTagItemType(argv[1]);
+	struct locate_item_list *conditionals;
+	int tagType = locate_parse_type(argv[1]);
 	int ret;
 
 	if (tagType < 0) {
@@ -992,28 +1199,28 @@ handle_list(struct client *client, int argc, char *argv[])
 		if (tagType != TAG_ITEM_ALBUM) {
 			command_error(client, ACK_ERROR_ARG,
 				      "should be \"%s\" for 3 arguments",
-				      mpdTagItemKeys[TAG_ITEM_ALBUM]);
+				      tag_item_names[TAG_ITEM_ALBUM]);
 			return COMMAND_RETURN_ERROR;
 		}
-		conditionals = newLocateTagItem(mpdTagItemKeys[TAG_ITEM_ARTIST],
-						argv[2]);
-		numConditionals = 1;
-	} else {
-		numConditionals =
-		    newLocateTagItemArrayFromArgArray(argv + 2,
-						      argc - 2, &conditionals);
 
-		if (numConditionals < 0) {
+		locate_item_list_parse(argv + 1, argc - 1);
+
+		conditionals = locate_item_list_new(1);
+		conditionals->items[0].tag = TAG_ITEM_ARTIST;
+		conditionals->items[0].needle = g_strdup(argv[2]);
+	} else {
+		conditionals =
+			locate_item_list_parse(argv + 2, argc - 2);
+		if (conditionals == NULL) {
 			command_error(client, ACK_ERROR_ARG,
 				      "not able to parse args");
 			return COMMAND_RETURN_ERROR;
 		}
 	}
 
-	ret = listAllUniqueTags(client, tagType, numConditionals, conditionals);
+	ret = listAllUniqueTags(client, tagType, conditionals);
 
-	if (conditionals)
-		freeLocateTagItemArray(numConditionals, conditionals);
+	locate_item_list_free(conditionals);
 
 	if (ret == -1)
 		command_error(client, ACK_ERROR_NO_EXIST,
@@ -1025,14 +1232,16 @@ handle_list(struct client *client, int argc, char *argv[])
 static enum command_return
 handle_move(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
-	int from, to;
+	unsigned start, end;
+	int to;
 	enum playlist_result result;
 
-	if (!check_int(client, &from, argv[1], check_integer, argv[1]))
+	if (!check_range(client, &start, &end,
+				      argv[1], need_range))
 		return COMMAND_RETURN_ERROR;
 	if (!check_int(client, &to, argv[2], check_integer, argv[2]))
 		return COMMAND_RETURN_ERROR;
-	result = moveSongInPlaylist(from, to);
+	result = moveSongRangeInPlaylist(&g_playlist, start, end, to);
 	return print_playlist_result(client, result);
 }
 
@@ -1046,7 +1255,7 @@ handle_moveid(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 		return COMMAND_RETURN_ERROR;
 	if (!check_int(client, &to, argv[2], check_integer, argv[2]))
 		return COMMAND_RETURN_ERROR;
-	result = moveSongInPlaylistById(id, to);
+	result = moveSongInPlaylistById(&g_playlist, id, to);
 	return print_playlist_result(client, result);
 }
 
@@ -1060,7 +1269,7 @@ handle_swap(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 		return COMMAND_RETURN_ERROR;
 	if (!check_int(client, &song2, argv[2], check_integer, argv[2]))
 		return COMMAND_RETURN_ERROR;
-	result = swapSongsInPlaylist(song1, song2);
+	result = swapSongsInPlaylist(&g_playlist, song1, song2);
 	return print_playlist_result(client, result);
 }
 
@@ -1074,7 +1283,7 @@ handle_swapid(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 		return COMMAND_RETURN_ERROR;
 	if (!check_int(client, &id2, argv[2], check_integer, argv[2]))
 		return COMMAND_RETURN_ERROR;
-	result = swapSongsInPlaylistById(id1, id2);
+	result = swapSongsInPlaylistById(&g_playlist, id1, id2);
 	return print_playlist_result(client, result);
 }
 
@@ -1089,7 +1298,7 @@ handle_seek(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 	if (!check_int(client, &seek_time, argv[2], check_integer, argv[2]))
 		return COMMAND_RETURN_ERROR;
 
-	result = seekSongInPlaylist(song, seek_time);
+	result = seekSongInPlaylist(&g_playlist, song, seek_time);
 	return print_playlist_result(client, result);
 }
 
@@ -1104,7 +1313,7 @@ handle_seekid(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 	if (!check_int(client, &seek_time, argv[2], check_integer, argv[2]))
 		return COMMAND_RETURN_ERROR;
 
-	result = seekSongInPlaylistById(id, seek_time);
+	result = seekSongInPlaylistById(&g_playlist, id, seek_time);
 	return print_playlist_result(client, result);
 }
 
@@ -1163,34 +1372,38 @@ static enum command_return
 handle_enableoutput(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
 	unsigned device;
-	int ret;
+	bool ret;
 
 	if (!check_unsigned(client, &device, argv[1]))
 		return COMMAND_RETURN_ERROR;
 
-	ret = enableAudioDevice(device);
-	if (ret == -1)
+	ret = audio_output_enable_index(device);
+	if (!ret) {
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "No such audio output");
+		return COMMAND_RETURN_ERROR;
+	}
 
-	return ret;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
 handle_disableoutput(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
 	unsigned device;
-	int ret;
+	bool ret;
 
 	if (!check_unsigned(client, &device, argv[1]))
 		return COMMAND_RETURN_ERROR;
 
-	ret = disableAudioDevice(device);
-	if (ret == -1)
+	ret = audio_output_disable_index(device);
+	if (!ret) {
 		command_error(client, ACK_ERROR_NO_EXIST,
 			      "No such audio output");
+		return COMMAND_RETURN_ERROR;
+	}
 
-	return ret;
+	return COMMAND_RETURN_OK;
 }
 
 static enum command_return
@@ -1224,17 +1437,19 @@ static enum command_return
 handle_playlistadd(struct client *client, G_GNUC_UNUSED int argc, char *argv[])
 {
 	char *playlist = argv[1];
-	char *path = argv[2];
+	char *uri = argv[2];
 	enum playlist_result result;
 
-	if (isRemoteUrl(path))
-		result = spl_append_uri(path, playlist);
-	else if (uri_has_scheme(path)) {
-		command_error(client, ACK_ERROR_NO_EXIST,
-			      "unsupported URI scheme");
-		return COMMAND_RETURN_ERROR;
+	if (uri_has_scheme(uri)) {
+		if (!uri_supported_scheme(uri)) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "unsupported URI scheme");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		result = spl_append_uri(uri, playlist);
 	} else
-		result = addAllInToStoredPlaylist(path, playlist);
+		result = addAllInToStoredPlaylist(uri, playlist);
 
 	if (result == (enum playlist_result)-1) {
 		command_error(client, ACK_ERROR_NO_EXIST,
@@ -1275,7 +1490,7 @@ handle_idle(struct client *client,
                         continue;
 
                 for (j = 0; idle_names[j]; ++j) {
-                        if (!strcasecmp(argv[i], idle_names[j])) {
+                        if (!g_ascii_strcasecmp(argv[i], idle_names[j])) {
                                 flags |= (1 << j);
                         }
                 }
@@ -1292,6 +1507,162 @@ handle_idle(struct client *client,
 	return 1;
 }
 
+#ifdef ENABLE_SQLITE
+struct sticker_song_find_data {
+	struct client *client;
+	const char *name;
+};
+
+static void
+sticker_song_find_print_cb(struct song *song, const char *value,
+			   gpointer user_data)
+{
+	struct sticker_song_find_data *data = user_data;
+
+	song_print_url(data->client, song);
+	sticker_print_value(data->client, data->name, value);
+}
+
+static enum command_return
+handle_sticker_song(struct client *client, int argc, char *argv[])
+{
+	if (argc == 5 && strcmp(argv[1], "get") == 0) {
+		struct song *song;
+		char *value;
+
+		song = db_get_song(argv[3]);
+		if (song == NULL) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no such song");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		value = sticker_song_get_value(song, argv[4]);
+		if (value == NULL) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no such sticker");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		sticker_print_value(client, argv[4], value);
+		g_free(value);
+
+		return COMMAND_RETURN_OK;
+	} else if (argc == 4 && strcmp(argv[1], "list") == 0) {
+		struct song *song;
+		struct sticker *sticker;
+
+		song = db_get_song(argv[3]);
+		if (song == NULL) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no such song");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		sticker = sticker_song_get(song);
+		if (NULL == sticker) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no stickers found");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		sticker_print(client, sticker);
+		sticker_free(sticker);
+
+		return COMMAND_RETURN_OK;
+	} else if (argc == 6 && strcmp(argv[1], "set") == 0) {
+		struct song *song;
+		bool ret;
+
+		song = db_get_song(argv[3]);
+		if (song == NULL) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no such song");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		ret = sticker_song_set_value(song, argv[4], argv[5]);
+		if (!ret) {
+			command_error(client, ACK_ERROR_SYSTEM,
+				      "failed to set sticker value");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		return COMMAND_RETURN_OK;
+	} else if ((argc == 4 || argc == 5) &&
+		   strcmp(argv[1], "delete") == 0) {
+		struct song *song;
+		bool ret;
+
+		song = db_get_song(argv[3]);
+		if (song == NULL) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no such song");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		ret = argc == 4
+			? sticker_song_delete(song)
+			: sticker_song_delete_value(song, argv[4]);
+		if (!ret) {
+			command_error(client, ACK_ERROR_SYSTEM,
+				      "no such sticker");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		return COMMAND_RETURN_OK;
+	} else if (argc == 5 && strcmp(argv[1], "find") == 0) {
+		/* "sticker find song a/directory name" */
+		struct directory *directory;
+		bool success;
+		struct sticker_song_find_data data = {
+			.client = client,
+			.name = argv[4],
+		};
+
+		directory = db_get_directory(argv[3]);
+		if (directory == NULL) {
+			command_error(client, ACK_ERROR_NO_EXIST,
+				      "no such directory");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		success = sticker_song_find(directory, data.name,
+					    sticker_song_find_print_cb, &data);
+		if (!success) {
+			command_error(client, ACK_ERROR_SYSTEM,
+				      "failed to set search sticker database");
+			return COMMAND_RETURN_ERROR;
+		}
+
+		return COMMAND_RETURN_OK;
+	} else {
+		command_error(client, ACK_ERROR_ARG, "bad request");
+		return COMMAND_RETURN_ERROR;
+	}
+}
+
+static enum command_return
+handle_sticker(struct client *client, int argc, char *argv[])
+{
+	assert(argc >= 4);
+
+	if (!sticker_enabled()) {
+		command_error(client, ACK_ERROR_UNKNOWN,
+			      "sticker database is disabled");
+		return COMMAND_RETURN_ERROR;
+	}
+
+	if (strcmp(argv[2], "song") == 0)
+		return handle_sticker_song(client, argc, argv);
+	else {
+		command_error(client, ACK_ERROR_ARG,
+			      "unknown sticker domain");
+		return COMMAND_RETURN_ERROR;
+	}
+}
+#endif
+
 /**
  * The command registry.
  *
@@ -1304,6 +1675,7 @@ static const struct command commands[] = {
 	{ "clearerror", PERMISSION_CONTROL, 0, 0, handle_clearerror },
 	{ "close", PERMISSION_NONE, -1, -1, handle_close },
 	{ "commands", PERMISSION_NONE, 0, 0, handle_commands },
+	{ "consume", PERMISSION_CONTROL, 1, 1, handle_consume },
 	{ "count", PERMISSION_READ, 2, -1, handle_count },
 	{ "crossfade", PERMISSION_CONTROL, 1, 1, handle_crossfade },
 	{ "currentsong", PERMISSION_READ, 0, 0, handle_currentsong },
@@ -1353,9 +1725,13 @@ static const struct command commands[] = {
 	{ "seek", PERMISSION_CONTROL, 2, 2, handle_seek },
 	{ "seekid", PERMISSION_CONTROL, 2, 2, handle_seekid },
 	{ "setvol", PERMISSION_CONTROL, 1, 1, handle_setvol },
-	{ "shuffle", PERMISSION_CONTROL, 0, 0, handle_shuffle },
+	{ "shuffle", PERMISSION_CONTROL, 0, 1, handle_shuffle },
+	{ "single", PERMISSION_CONTROL, 1, 1, handle_single },
 	{ "stats", PERMISSION_READ, 0, 0, handle_stats },
 	{ "status", PERMISSION_READ, 0, 0, handle_status },
+#ifdef ENABLE_SQLITE
+	{ "sticker", PERMISSION_ADMIN, 3, -1, handle_sticker },
+#endif
 	{ "stop", PERMISSION_CONTROL, 0, 0, handle_stop },
 	{ "swap", PERMISSION_CONTROL, 2, 2, handle_swap },
 	{ "swapid", PERMISSION_CONTROL, 2, 2, handle_swapid },
@@ -1367,6 +1743,17 @@ static const struct command commands[] = {
 
 static const unsigned num_commands = sizeof(commands) / sizeof(commands[0]);
 
+static bool
+command_available(G_GNUC_UNUSED const struct command *cmd)
+{
+#ifdef ENABLE_SQLITE
+	if (strcmp(cmd->cmd, "sticker") == 0)
+		return sticker_enabled();
+#endif
+
+	return true;
+}
+
 /* don't be fooled, this is the command handler for "commands" command */
 static enum command_return
 handle_commands(struct client *client,
@@ -1378,7 +1765,8 @@ handle_commands(struct client *client,
 	for (unsigned i = 0; i < num_commands; ++i) {
 		cmd = &commands[i];
 
-		if (cmd->permission == (permission & cmd->permission))
+		if (cmd->permission == (permission & cmd->permission) &&
+		    command_available(cmd))
 			client_printf(client, "command: %s\n", cmd->cmd);
 	}
 
@@ -1535,10 +1923,10 @@ command_process_list(struct client *client,
 	for (GSList *cur = list; cur != NULL; cur = g_slist_next(cur)) {
 		char *cmd = cur->data;
 
-		DEBUG("command_process_list: process command \"%s\"\n",
-		      cmd);
+		g_debug("command_process_list: process command \"%s\"",
+			cmd);
 		ret = command_process(client, cmd);
-		DEBUG("command_process_list: command returned %i\n", ret);
+		g_debug("command_process_list: command returned %i", ret);
 		if (ret != COMMAND_RETURN_OK || client_is_expired(client))
 			break;
 		else if (list_ok)
