@@ -55,7 +55,7 @@ struct ffmpeg_context {
 
 struct ffmpeg_stream {
 	/** hack - see url_to_struct() */
-	char url[8];
+	char url[64];
 
 	struct decoder *decoder;
 	struct input_stream *input;
@@ -140,8 +140,28 @@ ffmpeg_find_audio_stream(const AVFormatContext *format_context)
 	return -1;
 }
 
+/**
+ * Append the suffix of the original URI to the virtual stream URI.
+ * Without this, libavformat cannot detect some of the codecs
+ * (e.g. "shorten").
+ */
+static void
+append_uri_suffix(struct ffmpeg_stream *stream, const char *uri)
+{
+	assert(stream != NULL);
+	assert(uri != NULL);
+
+	char *base = g_path_get_basename(uri);
+
+	const char *suffix = strrchr(base, '.');
+	if (suffix != NULL && suffix[1] != 0)
+		g_strlcat(stream->url, suffix, sizeof(stream->url));
+
+	g_free(base);
+}
+
 static bool
-ffmpeg_helper(struct input_stream *input,
+ffmpeg_helper(const char *uri, struct input_stream *input,
 	      bool (*callback)(struct ffmpeg_context *ctx),
 	      struct ffmpeg_context *ctx)
 {
@@ -154,6 +174,9 @@ ffmpeg_helper(struct input_stream *input,
 	};
 	bool ret;
 
+	if (uri != NULL)
+		append_uri_suffix(&stream, uri);
+
 	stream.input = input;
 	if (ctx && ctx->decoder) {
 		stream.decoder = ctx->decoder; //are we in decoding loop ?
@@ -209,6 +232,21 @@ ffmpeg_helper(struct input_stream *input,
 	return ret;
 }
 
+/**
+ * On some platforms, libavcodec wants the output buffer aligned to 16
+ * bytes (because it uses SSE/Altivec internally).  This function
+ * returns the aligned version of the specified buffer, and corrects
+ * the buffer size.
+ */
+static void *
+align16(void *p, size_t *length_p)
+{
+	unsigned add = 16 - (size_t)p % 16;
+
+	*length_p -= add;
+	return (char *)p + add;
+}
+
 static enum decoder_command
 ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 		   const AVPacket *packet,
@@ -217,7 +255,9 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 {
 	enum decoder_command cmd = DECODE_COMMAND_NONE;
 	int position;
-	uint8_t audio_buf[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2];
+	uint8_t audio_buf[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2 + 16];
+	int16_t *aligned_buffer;
+	size_t buffer_size;
 	int len, audio_size;
 	uint8_t *packet_data;
 	int packet_size;
@@ -225,11 +265,13 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 	packet_data = packet->data;
 	packet_size = packet->size;
 
+	buffer_size = sizeof(audio_buf);
+	aligned_buffer = align16(audio_buf, &buffer_size);
+
 	while ((packet_size > 0) && (cmd == DECODE_COMMAND_NONE)) {
-		audio_size = sizeof(audio_buf);
+		audio_size = buffer_size;
 		len = avcodec_decode_audio2(codec_context,
-					    (int16_t *)audio_buf,
-					    &audio_size,
+					    aligned_buffer, &audio_size,
 					    packet_data, packet_size);
 
 		if (len < 0) {
@@ -250,7 +292,7 @@ ffmpeg_send_packet(struct decoder *decoder, struct input_stream *is,
 			: 0;
 
 		cmd = decoder_data(decoder, is,
-				   audio_buf, audio_size,
+				   aligned_buffer, audio_size,
 				   position,
 				   codec_context->bit_rate / 1000, NULL);
 	}
@@ -270,10 +312,6 @@ ffmpeg_decode_internal(struct ffmpeg_context *ctx)
 
 	total_time = 0;
 
-	if (codec_context->channels > 2) {
-		codec_context->channels = 2;
-	}
-
 #if LIBAVCODEC_VERSION_INT >= ((51<<16)+(41<<8)+0)
 	audio_format.bits = (uint8_t) av_get_bits_per_sample_format(codec_context->sample_fmt);
 #else
@@ -334,7 +372,8 @@ ffmpeg_decode(struct decoder *decoder, struct input_stream *input)
 	ctx.input = input;
 	ctx.decoder = decoder;
 
-	ffmpeg_helper(input, ffmpeg_decode_internal, &ctx);
+	ffmpeg_helper(decoder_get_uri(decoder), input,
+		      ffmpeg_decode_internal, &ctx);
 }
 
 #if LIBAVFORMAT_VERSION_INT >= ((52<<16)+(31<<8)+0)
@@ -352,17 +391,17 @@ ffmpeg_copy_metadata(struct tag *tag, AVMetadata *m,
 static bool ffmpeg_tag_internal(struct ffmpeg_context *ctx)
 {
 	struct tag *tag = (struct tag *) ctx->tag;
-	const AVFormatContext *f = ctx->format_context;
+	AVFormatContext *f = ctx->format_context;
 
 	tag->time = 0;
 	if (f->duration != (int64_t)AV_NOPTS_VALUE)
 		tag->time = f->duration / AV_TIME_BASE;
 
 #if LIBAVFORMAT_VERSION_INT >= ((52<<16)+(31<<8)+0)
+	av_metadata_conv(f, NULL, f->iformat->metadata_conv);
+
 	ffmpeg_copy_metadata(tag, f->metadata, TAG_ITEM_TITLE, "title");
-	if (!ffmpeg_copy_metadata(tag, f->metadata, TAG_ITEM_ARTIST, "author"))
-		ffmpeg_copy_metadata(tag, f->metadata,
-				     TAG_ITEM_ARTIST, "artist");
+	ffmpeg_copy_metadata(tag, f->metadata, TAG_ITEM_ARTIST, "author");
 	ffmpeg_copy_metadata(tag, f->metadata, TAG_ITEM_ALBUM, "album");
 	ffmpeg_copy_metadata(tag, f->metadata, TAG_ITEM_COMMENT, "comment");
 	ffmpeg_copy_metadata(tag, f->metadata, TAG_ITEM_GENRE, "genre");
@@ -411,7 +450,7 @@ static struct tag *ffmpeg_tag(const char *file)
 	ctx.decoder = NULL;
 	ctx.tag = tag_new();
 
-	ret = ffmpeg_helper(&input, ffmpeg_tag_internal, &ctx);
+	ret = ffmpeg_helper(file, &input, ffmpeg_tag_internal, &ctx);
 	if (!ret) {
 		tag_free(ctx.tag);
 		ctx.tag = NULL;
@@ -447,25 +486,31 @@ static const char *const ffmpeg_suffixes[] = {
 };
 
 static const char *const ffmpeg_mime_types[] = {
+	"application/m4a",
 	"application/mp4",
 	"application/octet-stream",
 	"application/ogg",
 	"application/x-ms-wmz",
 	"application/x-ms-wmd",
+	"application/x-ogg",
 	"application/x-shockwave-flash",
 	"application/x-shorten",
 	"audio/8svx",
 	"audio/16sv",
 	"audio/aac",
 	"audio/ac3",
+	"audio/aiff"
 	"audio/amr",
 	"audio/basic",
 	"audio/flac",
+	"audio/m4a",
+	"audio/mp4",
 	"audio/mpeg",
 	"audio/musepack",
 	"audio/ogg",
 	"audio/qcelp",
 	"audio/vorbis",
+	"audio/vorbis+ogg",
 	"audio/x-8svx",
 	"audio/x-16sv",
 	"audio/x-aac",
@@ -478,15 +523,20 @@ static const char *const ffmpeg_mime_types[] = {
 	"audio/x-flac",
 	"audio/x-gsm",
 	"audio/x-mace",
+	"audio/x-matroska",
 	"audio/x-monkeys-audio",
 	"audio/x-mpeg",
 	"audio/x-ms-wma",
 	"audio/x-ms-wax",
 	"audio/x-musepack",
+	"audio/x-ogg",
+	"audio/x-vorbis",
+	"audio/x-vorbis+ogg",
 	"audio/x-pn-realaudio",
 	"audio/x-pn-multirate-realaudio",
 	"audio/x-speex",
 	"audio/x-tta"
+	"audio/x-voc",
 	"audio/x-wav",
 	"audio/x-wma",
 	"audio/x-wv",
