@@ -272,6 +272,10 @@ oggflac_tag_dup(const char *file)
 		return NULL;
 	}
 
+	/* rewind the stream, because ogg_stream_type_detect() has
+	   moved it */
+	input_stream_seek(&input_stream, 0, SEEK_SET);
+
 	flac_data_init(&data, NULL, &input_stream);
 
 	data.tag = tag_new();
@@ -300,6 +304,10 @@ oggflac_decode(struct decoder * mpd_decoder, struct input_stream *input_stream)
 	if (ogg_stream_type_detect(input_stream) != FLAC)
 		return;
 
+	/* rewind the stream, because ogg_stream_type_detect() has
+	   moved it */
+	input_stream_seek(input_stream, 0, SEEK_SET);
+
 	flac_data_init(&data, mpd_decoder, input_stream);
 
 	if (!(decoder = full_decoder_init_and_read_metadata(&data, 0))) {
@@ -349,9 +357,11 @@ fail:
 
 static const char *const oggflac_suffixes[] = { "ogg", "oga", NULL };
 static const char *const oggflac_mime_types[] = {
-	"audio/x-flac+ogg",
 	"application/ogg",
 	"application/x-ogg",
+	"audio/ogg",
+	"audio/x-ogg",
+	"audio/x-flac+ogg",
 	NULL
 };
 
